from AOT_biomaps.Settings import Params
from AOT_biomaps.AOT_Optic._mainOptic import Phantom
from AOT_biomaps.AOT_Acoustic.AcousticEnums import WaveType, FormatSave
from AOT_biomaps.AOT_Acoustic.StructuredWave import StructuredWave

from abc import ABC, abstractmethod
import os
import numpy as np
import torch
import torch.nn.functional as F
from tqdm import trange
from datetime import datetime
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib as mpl
import copy

class Experiment(ABC):
    def __init__(self, params, acousticType=WaveType.StructuredWave, formatSave=FormatSave.HDR_IMG):
        self.params = params
        self.OpticImage = None
        self.AcousticFields = None
        self.AOsignal_withTumor = None
        self.AOsignal_withoutTumor = None

        if type(acousticType).__name__ != "WaveType":
            raise TypeError("acousticType must be an instance of the WaveType class")

        self.FormatSave = formatSave
        self.TypeAcoustic = acousticType

        if type(self.params) != Params:
            raise TypeError("params must be an instance of the Params class")

    def copy(self):
        """Retourne une copie profonde de l'objet."""
        return copy.deepcopy(self)
    
    def generatePhantom(self):
        """
        Generate the phantom for the experiment.
        This method initializes the OpticImage attribute with a Phantom instance.
        """
        self.OpticImage = Phantom(params=self.params)

    @abstractmethod
    def generateAcousticFields(self, fieldDataPath, fieldParamPath, show_log=True):
        """
        Generate the acoustic fields for simulation.
        Args:
            fieldDataPath: Path to save the generated fields.
            fieldParamPath: Path to the field parameters file.
        Returns:
            systemMatrix: A numpy array of the generated fields.
        """
        pass

    def cutAcousticFields(self, max_t, min_t=0):

        max_t = float(max_t)
        min_t = float(min_t)

        min_sample = int(np.floor(min_t * float(self.params.acoustic['f_saving'])))
        max_sample = int(np.floor(max_t * float(self.params.acoustic['f_saving'])))

        if min_sample < 0 or max_sample < 0:
            raise ValueError("min_sample and max_sample must be non-negative integers.")
        if min_sample >= max_sample:
            raise ValueError("min_sample must be less than max_sample.")

        if not self.AcousticFields:
            raise ValueError("AcousticFields is empty. Cannot cut fields.")

        for i in trange(len(self.AcousticFields), desc=f"Cutting Acoustic Fields ({min_sample} to {max_sample} samples)"):
            field = self.AcousticFields[i]
            if field.field.shape[0] < max_sample:
                raise ValueError(f"Field {field.getName_field()} has an invalid shape: {field.field.shape}. Expected shape to be at least ({max_sample},).")
            self.AcousticFields[i].field = field.field[min_sample:max_sample, :, :]

    def addNoise(self, noiseType='gaussian', noiseLvl=0.1, withTumor=True):
        """
        Ajoute du bruit (gaussien ou poisson) au signal AO sélectionné.

        Args:
            noiseType (str): Type de bruit à ajouter ('gaussian' ou 'poisson').
            noiseLvl (float): Niveau de bruit (écart-type pour le bruit gaussien, facteur multiplicatif pour le bruit de Poisson).
            withTumor (bool): Si True, ajoute le bruit au signal avec tumeur, sinon au signal sans tumeur.
        """
        if withTumor and self.AOsignal_withTumor is None:
            raise ValueError("AO signal with tumor is not generated. Please generate it first.")
        if not withTumor and self.AOsignal_withoutTumor is None:
            raise ValueError("AO signal without tumor is not generated. Please generate it first.")

        if withTumor:
            AOsignals = self.AOsignal_withTumor
        else:
            AOsignals = self.AOsignal_withoutTumor
        
        noiseSignals = np.zeros_like(AOsignals)
        for i in trange(AOsignals.shape[1], desc=f"Adding {noiseType} noise to AO signal {'with' if withTumor else 'without'} tumor"):
            AOsignal = AOsignals[:, i]
            if noiseType.lower() == 'gaussian':
                noise = np.random.normal(0, noiseLvl*np.max(AOsignal), AOsignal.shape)
                noisy_signal = AOsignal + noise
            elif noiseType.lower() == 'poisson':
                # Pour le bruit de Poisson, on utilise souvent un facteur multiplicatif
                # car le bruit de Poisson est proportionnel à la racine carrée du signal.
                # Ici, on multiplie le signal par un facteur aléatoire centré autour de 1.
                noise = np.random.poisson(noiseLvl * np.abs(AOsignal)) / (noiseLvl * np.abs(AOsignal).max())
                noisy_signal = AOsignal * noise
            else:
                raise ValueError("noiseType must be either 'gaussian' or 'poisson'.")
            noisy_signal = np.clip(noisy_signal, a_min=0, a_max=None)  # Assurer que le signal reste non négatif
            noiseSignals[:, i] = noisy_signal
        return noiseSignals

    def reduceDims(self, mode='avg'):
        """
        Réduit les dimensions T, X, Z d'un numpy array (T, X, Z) par 2 en utilisant une convolution.
        Retourne un numpy array et met à jour les paramètres numériques.
        """
        for i in trange(len(self.AcousticFields),
                        desc="Downsampling Acoustic Fields (T, X, Z → T//2, X//2, Z//2)"):
            # Conversion en tenseur PyTorch
            field = self.AcousticFields[i].field
            if not isinstance(field, torch.Tensor):
                field = torch.from_numpy(field)

            # Vérification de la forme (doit être 3D : T, X, Z)
            if field.dim() != 3:
                raise ValueError(f"Forme non supportée : {field.shape}. Attendu (T, X, Z).")

            # Ajout des dimensions pour conv3d : (1, 1, T, X, Z)
            x = field.unsqueeze(0).unsqueeze(0)

            # Réduction par convolution 3D
            if mode == 'avg':
                x_down = F.avg_pool3d(x, kernel_size=(2, 2, 2), stride=(2, 2, 2))
            else:  # mode == 'max'
                x_down = F.max_pool3d(x, kernel_size=(2, 2, 2), stride=(2, 2, 2))

            # Conversion en numpy array et suppression des dimensions ajoutées
            self.AcousticFields[i].field = x_down.squeeze(0).squeeze(0).cpu().numpy()

        # Fonction utilitaire pour convertir et mettre à jour un paramètre
        def convert_and_update(param_dict, key, operation):
            if key in param_dict:
                if isinstance(param_dict[key], str):
                    param_dict[key] = float(param_dict[key])
                param_dict[key] = operation(param_dict[key])

        # Mise à jour des paramètres
        convert_and_update(self.params.acoustic, 'f_saving', lambda x: x / 2)
        for param in ['dx', 'dy', 'dz']:
            convert_and_update(self.params.general, param, lambda x: x * 2)

    def normalizeAOsignals(self, withTumor=True):
        if withTumor and self.AOsignal_withTumor is None:
            raise ValueError("AO signal with tumor is not generated. Please generate it first.")
        if not withTumor and self.AOsignal_withoutTumor is None:
            raise ValueError("AO signal without tumor is not generated. Please generate it first.")
        if withTumor:
            self.AOsignal_withTumor = self.AOsignal_withTumor - np.min(self.AOsignal_withTumor)/(np.max(self.AOsignal_withTumor)-np.min(self.AOsignal_withTumor))
        else:
            self.AOsignal_withoutTumor = self.AOsignal_withoutTumor - np.min(self.AOsignal_withoutTumor)/(np.max(self.AOsignal_withoutTumor)-np.min(self.AOsignal_withoutTumor))

    def saveAcousticFields(self, save_directory):
        progress_bar = trange(len(self.AcousticFields), desc="Saving Acoustic Fields")
        for i in progress_bar:
            progress_bar.set_postfix_str(f"-- {self.AcousticFields[i].getName_field()}")
            self.AcousticFields[i].save_field(save_directory, formatSave=self.FormatSave)

    def show_animated_Acoustic(self, wave_name=None, desired_duration_ms=5000, save_dir=None):
        """
        Plot synchronized animations of A_matrix slices for selected angles.
        Args:
            wave_name: optional name for labeling the subplots (e.g., "wave1")
            desired_duration_ms: Total duration of the animation in milliseconds.
            save_dir: directory to save the animation gif; if None, animation will not be saved
        Returns:
            ani: Matplotlib FuncAnimation object
        """
        mpl.rcParams['animation.embed_limit'] = 100
        if save_dir is not None:
            os.makedirs(save_dir, exist_ok=True)

        num_plots = len(self.AcousticFields)
        if num_plots <= 5:
            nrows, ncols = 1, num_plots
        else:
            ncols = 5
            nrows = (num_plots + ncols - 1) // ncols

        fig, axes = plt.subplots(nrows, ncols, figsize=(5 * ncols, 5.3 * nrows))
        if isinstance(axes, plt.Axes):
            axes = np.array([axes])
        axes = axes.flatten()
        ims = []

        fig.suptitle(f"System Matrix Animation {wave_name}", fontsize=12, y=0.98)

        for idx in range(num_plots):
            ax = axes[idx]
            im = ax.imshow(self.AcousticFields[0, :, :, idx],
                        extent=(self.params['Xrange'][0], self.params['Xrange'][1], self.params['Zrange'][1], self.params['Zrange'][0]),
                        vmax=1, aspect='equal', cmap='jet', animated=True)
            ax.set_xlabel("x (mm)", fontsize=8)
            ax.set_ylabel("z (mm)", fontsize=8)
            ims.append((im, ax, idx))

        for j in range(num_plots, len(axes)):
            fig.delaxes(axes[j])

        plt.tight_layout(rect=[0, 0, 1, 0.95])

        def update(frame):
            artists = []
            for im, ax, idx in ims:
                im.set_array(self.AcousticFields[frame, :, :, idx])
                fig.suptitle(f"System Matrix Animation {wave_name} t = {frame * 25e-6 * 1000:.2f} ms", fontsize=10)
                artists.append(im)
            return artists

        interval = desired_duration_ms / self.AcousticFields.shape[0]
        ani = animation.FuncAnimation(
            fig, update,
            frames=range(0, self.AcousticFields.shape[0]),
            interval=interval, blit=True
        )

        if save_dir is not None:
            now = datetime.now()
            date_str = now.strftime("%Y_%d_%m_%y")
            save_filename = f"AcousticField_{wave_name}_{date_str}.gif"
            save_path = os.path.join(save_dir, save_filename)
            ani.save(save_path, writer='pillow', fps=20)
            print(f"Saved: {save_path}")

        plt.close(fig)
        return ani

    def generateAOsignal(self, withTumor=True, AOsignalDataPath=None):

        if AOsignalDataPath is not None:
            if not os.path.exists(AOsignalDataPath):
                raise FileNotFoundError(f"AO file {AOsignalDataPath} not found.")
            if withTumor:
                self.AOsignal_withTumor = self._loadAOSignal(AOsignalDataPath)
                if self.AOsignal_withTumor.shape[0] != self.AcousticFields[0].field.shape[0]:
                    print(f"AO signal shape {self.AOsignal_withTumor.shape} does not match the expected shape {self.AcousticFields[0].field.shape}. Resizing Acoustic fields...")
                    self.cutAcousticFields(max_t=self.AOsignal_withTumor.shape[0] / float(self.params.acoustic['f_saving']), min_t=0)
            else:
                self.AOsignal_withoutTumor = self._loadAOSignal(AOsignalDataPath)
                if self.AOsignal_withoutTumor.shape[0] != self.AcousticFields[0].field.shape[0]:
                    print(f"AO signal shape {self.AOsignal_withoutTumor.shape} does not match the expected shape {self.AcousticFields[0].field.shape}. Resizing Acoustic fields...")
                    self.cutAcousticFields(max_t=self.AOsignal_withoutTumor.shape[0] / float(self.params.acoustic['f_saving']), min_t=0)
        else:    
            if self.AcousticFields is None:
                raise ValueError("AcousticFields is not initialized. Please generate the system matrix first.")

            if self.OpticImage is None:
                raise ValueError("OpticImage is not initialized. Please generate the phantom first.")
            
            if not all(field.field.shape == self.AcousticFields[0].field.shape for field in self.AcousticFields):
                minShape = min([field.field.shape[0] for field in self.AcousticFields])
                self.cutAcousticFields(max_t=minShape * self.params['fs_aq'])
            else:
                shape_field = self.AcousticFields[0].field.shape

            AOsignal = np.zeros((shape_field[0], len(self.AcousticFields)), dtype=np.float32)

            if withTumor:
                description = "Generating AO Signal with Tumor"
            else:
                description = "Generating AO Signal without Tumor"

            for i in trange(len(self.AcousticFields), desc=description):
                for t in range(self.AcousticFields[i].field.shape[0]):
                    if withTumor:
                        interaction = self.OpticImage.phantom * self.AcousticFields[i].field[t, :, :]
                    else:
                        interaction = self.OpticImage.laser.intensity * self.AcousticFields[i].field[t, :, :]
                    AOsignal[t, i] = np.sum(interaction)

            if withTumor:
                self.AOsignal_withTumor = AOsignal
            else:
                self.AOsignal_withoutTumor = AOsignal

    @staticmethod
    def _loadAOSignal(AOsignalPath):
        if AOsignalPath.endswith(".cdh"):
            with open(AOsignalPath, "r") as file:
                cdh_content = file.readlines()
            
            cdf_path = AOsignalPath.replace(".cdh", ".cdf")

            # Extraire les paramètres depuis le fichier .cdh
            n_scans = int([line.split(":")[1].strip() for line in cdh_content if "Number of events" in line][0])
            n_acquisitions_per_event = int([line.split(":")[1].strip() for line in cdh_content if "Number of acquisitions per event" in line][0])
            num_elements = int([line.split(":")[1].strip() for line in cdh_content if "Number of US transducers" in line][0])

            # Initialisation des structures
            AO_signal = np.zeros((n_acquisitions_per_event, n_scans), dtype=np.float32)
            active_lists = []
            angles = []

            # Lecture du fichier binaire
            with open(cdf_path, "rb") as file:
                for j in trange(n_scans, desc="Lecture des événements"):
                    # Lire l'activeList : 48 caractères hex = 24 bytes
                    active_list_bytes = file.read(24)
                    active_list_hex = active_list_bytes.hex()
                    active_lists.append(active_list_hex)

                    # Lire l'angle (1 byte signé)
                    angle_byte = file.read(1)
                    angle = np.frombuffer(angle_byte, dtype=np.int8)[0]
                    angles.append(angle)

                    # Lire le signal AO (float32)
                    data = np.frombuffer(file.read(n_acquisitions_per_event * 4), dtype=np.float32)
                    if len(data) != n_acquisitions_per_event:
                        raise ValueError(f"Erreur à l'événement {j} : attendu {n_acquisitions_per_event}, obtenu {len(data)}")
                    AO_signal[:, j] = data

            return AO_signal


        elif AOsignalPath.endswith(".npy"):
            return np.load(AOsignalPath)  # Supposé déjà au bon format
        else:
            raise ValueError("Format de fichier non supporté. Utilisez .cdh/.cdf ou .npy.")

    def saveAOsignals_Castor(self, save_directory, withTumor=True):
        if withTumor:
            AO_signal = self.AOsignal_withTumor
            cdf_location = os.path.join(save_directory, "AOSignals_withTumor.cdf")
            cdh_location = os.path.join(save_directory, "AOSignals_withTumor.cdh")
        else:
            AO_signal = self.AOsignal_withoutTumor
            cdf_location = os.path.join(save_directory, "AOSignals_withoutTumor.cdf")
            cdh_location = os.path.join(save_directory, "AOSignals_withoutTumor.cdh")

        info_location = os.path.join(save_directory, "info.txt")
        nScan = AO_signal.shape[1]

        with open(cdf_location, "wb") as fileID:
            for j in range(AO_signal.shape[1]):
                active_list_hex = self.AcousticFields[j].pattern.activeList
                for i in range(0, len(active_list_hex), 2):
                    byte_value = int(active_list_hex[i:i+2], 16)
                    fileID.write(byte_value.to_bytes(1, byteorder='big'))
                angle = self.AcousticFields[j].angle
                fileID.write(np.int8(angle).tobytes())
                fileID.write(AO_signal[:, j].astype(np.float32).tobytes())

        header_content = (
            f"Data filename: {'AOSignals_withTumor.cdf' if withTumor else 'AOSignals_withoutTumor.cdf'}\n"
            f"Number of events: {nScan}\n"
            f"Number of acquisitions per event: {AO_signal.shape[0]}\n"
            f"Start time (s): 0\n"
            f"Duration (s): 1\n"
            f"Acquisition frequency (Hz): {self.params.acoustic['f_saving']}\n"
            f"Data mode: histogram\n"
            f"Data type: AOT\n"
            f"Number of US transducers: {self.params.acoustic['num_elements']}"
        )

        with open(cdh_location, "w") as fileID:
            fileID.write(header_content)

        with open(info_location, "w") as fileID:
            for field in self.AcousticFields:
                fileID.write(field.getName_field() + "\n")

        print(f"Fichiers .cdf, .cdh et info.txt sauvegardés dans {save_directory}")

    def showAOsignal(self, withTumor=True, save_dir=None, wave_name=None):
        if withTumor and self.AOsignal_withTumor is None:
            raise ValueError("AO signal with tumor is not generated. Please generate it first.")
        if not withTumor and self.AOsignal_withoutTumor is None:
            raise ValueError("AO signal without tumor is not generated. Please generate it first.")

        if withTumor:
            AOsignal = self.AOsignal_withTumor
        else:
            AOsignal = self.AOsignal_withoutTumor

        time_axis = np.arange(AOsignal.shape[0]) / float(self.params.acoustic['f_AQ']) * 1e6

        num_plots = AOsignal.shape[1]
        if num_plots <= 5:
            nrows, ncols = 1, num_plots
        else:
            ncols = 5
            nrows = (num_plots + ncols - 1) // ncols

        fig, axes = plt.subplots(nrows, ncols, figsize=(5 * ncols, 5.3 * nrows))
        if isinstance(axes, plt.Axes):
            axes = np.array([axes])
        axes = axes.flatten()

        if wave_name is None:
            title = "AO Signal -- all plots"
        else:
            title = f"AO Signal -- {wave_name}"

        fig.suptitle(title, fontsize=12, y=0.98)

        for idx in range(num_plots):
            ax = axes[idx]
            ax.plot(time_axis, AOsignal[:, idx])
            ax.set_xlabel("Time (µs)", fontsize=8)
            ax.set_ylabel("Value", fontsize=8)

        for j in range(num_plots, len(axes)):
            fig.delaxes(axes[j])

        plt.tight_layout(rect=[0, 0, 1, 0.95])

        if save_dir is not None:
            now = datetime.now()
            date_str = now.strftime("%Y_%d_%m_%y")
            os.makedirs(save_dir, exist_ok=True)
            save_filename = f"Static_y_Plot{wave_name}_{date_str}.png"
            save_path = os.path.join(save_dir, save_filename)
            plt.savefig(save_path, dpi=200)
            print(f"Saved: {save_path}")

        plt.show()
        plt.close(fig)

    def show_animated_all(self, fileOfAcousticField=None, save_dir=None, desired_duration_ms=5000):
        mpl.rcParams['animation.embed_limit'] = 100
        pattern_str = StructuredWave.getPattern(fileOfAcousticField)
        angle = StructuredWave.getAngle(fileOfAcousticField)
        fieldToPlot = None

        for field in self.AcousticFields:
            if field.get_path() == fileOfAcousticField:
                fieldToPlot = field
                idx = self.AcousticFields.index(field)
                break
        else:
            raise ValueError(f"Field {fileOfAcousticField} not found in AcousticFields.")

        if wave_name is None:
            wave_name = f"Pattern structure {pattern_str}"

        fig, axs = plt.subplots(1, 2, figsize=(6 * 2, 5.3 * 1))
        if isinstance(axs, plt.Axes):
            axs = np.array([axs])

        fig.suptitle(f"AO Signal Animation {wave_name} | Angle {angle}°", fontsize=12, y=0.98)

        axs[0].imshow(self.OpticImage.T, cmap='hot', alpha=1, origin='upper',
                    extent=(self.params['Xrange'][0], self.params['Xrange'][1], self.params['Zrange'][1], self.params['Zrange'][0]),
                    aspect='equal')

        im_field = axs[0].imshow(fieldToPlot[0, :, :, idx], cmap='jet', origin='upper',
                                extent=(self.params['Xrange'][0], self.params['Xrange'][1], self.params['Zrange'][1], self.params['Zrange'][0]),
                                vmax=1, vmin=0.01, alpha=0.8, aspect='equal')

        axs[0].set_title(f"{wave_name} | Angle {angle}° | t = 0.00 ms", fontsize=10)
        axs[0].set_xlabel("x (mm)", fontsize=8)
        axs[0].set_ylabel("z (mm)", fontsize=8)

        time_axis = np.arange(self.AOsignal.shape[0]) * 25e-6 * 1000
        line_y, = axs[1].plot(time_axis, self.AOsignal[:, idx])
        vertical_line, = axs[1].plot([time_axis[0], time_axis[0]], [0, self.AOsignal[0, idx]], 'r--')

        axs[1].set_xlabel("Time (ms)", fontsize=8)
        axs[1].set_ylabel("Value", fontsize=8)
        axs[1].set_title(f"{wave_name} | Angle {angle}° | t = 0.00 ms", fontsize=10)

        plt.tight_layout(rect=[0, 0, 1, 0.95])

        def update(frame):
            current_time_ms = frame * 25e-6 * 1000
            frame_data = fieldToPlot[frame, :, :, idx]
            masked_data = np.where(frame_data > 0.02, frame_data, np.nan)
            im_field.set_data(masked_data)
            axs[0].set_title(f"{wave_name} | Angle {angle}° | t = {current_time_ms:.2f} ms", fontsize=10)

            y_vals = self.AOsignal[:, idx]
            y_copy = np.full_like(y_vals, np.nan)
            y_copy[:frame + 1] = y_vals[:frame + 1]
            line_y.set_data(time_axis, y_copy)

            vertical_line.set_data([time_axis[frame], time_axis[frame]], [0, y_vals[frame]])
            axs[1].set_title(f"{wave_name} | Angle {angle}° | t = {current_time_ms:.2f} ms", fontsize=10)

            return [im_field, vertical_line, line_y]

        interval = desired_duration_ms / fieldToPlot.shape[0]
        ani = animation.FuncAnimation(
            fig, update,
            frames=range(0, self.AcousticFields.shape[0]),
            interval=interval, blit=True
        )

        if save_dir is not None:
            now = datetime.now()
            date_str = now.strftime("%Y_%d_%m_%y")
            os.makedirs(save_dir, exist_ok=True)
            save_filename = f"A_y_LAMBDA_overlay_{pattern_str}_{angle}_{date_str}.gif"
            save_path = os.path.join(save_dir, save_filename)
            ani.save(save_path, writer='pillow', fps=20)
            print(f"Saved: {save_path}")

        plt.close(fig)
        return ani

    def showPhantom(self, withROI=False):
        """
        Displays the optical phantom with absorbers.
        """
        try:
            if withROI:
                self.OpticImage.show_ROI()
            else:
                self.OpticImage.show_phantom()
        except Exception as e:
            raise RuntimeError(f"Error plotting phantom: {e}")
        
    @abstractmethod
    def check(self):
        """
        Check if the experiment is correctly initialized.
        """
        pass



# sparse_matrix_gpu.py
import pycuda.driver as drv
import numpy as np
from pycuda.compiler import SourceModule
from tqdm import trange
import gc
import os

class SparseSMatrix_CSR:
    """Construction d'une matrice CSR à partir d'un objet `manip`.
    Usage:
        S = SparseMatrixGPU(manip)
        S.allocate()
    Après allocate(), on a: row_ptr (host np.int64 array), row_ptr_gpu (device ptr),
    h_col_ind, h_values, col_ind_gpu, values_gpu, norm_factor_inv.
    """

    def __init__(self, manip, block_rows=64, relative_threshold=0.3, device=0):
        drv.init()
        self.device = drv.Device(device)
        self.ctx = self.device.make_context()
        self.manip = manip
        self.N = len(manip.AcousticFields)
        self.T = manip.AcousticFields[0].field.shape[0]
        self.Z = manip.AcousticFields[0].field.shape[1]
        self.X = manip.AcousticFields[0].field.shape[2]
        self.block_rows = block_rows
        self.relative_threshold = relative_threshold
        self.h_dense = None
        self.row_ptr = None
        self.row_ptr_gpu = None
        self.h_col_ind = None
        self.h_values = None
        self.total_nnz = 0
        self.norm_factor_inv = None
        self.sparse_mod = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.free()

    def load_precompiled_module(self, so_path="AOT_biomaps_kernels.cubin"):
        try:
            # If a PTX or cubin is provided via path
            self.sparse_mod = drv.module_from_file(so_path)
            print(f"✅ Module CUDA chargé depuis {so_path}")
        except Exception:
            # Fallback: try to compile from bundled source (if available)
            src_path = os.path.join(os.path.dirname(__file__), 'AOT_biomaps_kernels.cu')
            if os.path.exists(src_path):
                print("Compilation JIT du kernel CUDA depuis source...")
                with open(src_path, 'r') as f:
                    src = f.read()
                self.sparse_mod = SourceModule(src, no_extern_c=True)
                print("✅ Module compilé JIT")
            else:
                raise

    def estimate_nnz_cpu(self):
        """Estimation rapide (non-exacte) — utile si tu veux une estimation faible.
        Recommandé : utiliser la passe GPU exacte (count_nnz_per_row_kernel) à la place.
        """
        total = 0
        for n in range(self.N):
            field = self.manip.AcousticFields[n].field
            for t in range(self.T):
                row = field[t].flatten()
                row_max = np.max(np.abs(row))
                thr = row_max * self.relative_threshold
                total += np.count_nonzero(np.abs(row) > thr)
        return int(total)

    def allocate(self, kernel_module_path=None):
        try:
            # --- 1. Construction bloc par bloc (sans garder tout le dense si possible) ---
            num_rows = self.N * self.T
            num_cols = self.Z * self.X
            bytes_float = np.dtype(np.float32).itemsize

            # Charge module
            if kernel_module_path:
                self.load_precompiled_module(kernel_module_path)
            else:
                self.load_precompiled_module('AOT_biomaps_kernels.cubin')

            count_nnz_kernel = self.sparse_mod.get_function('count_nnz_per_row_kernel')
            fill_csr_kernel = self.sparse_mod.get_function('fill_kernel__CSR')

            # allocate host row_ptr
            self.row_ptr = np.zeros(num_rows + 1, dtype=np.int64)

            # GPU temp buffers
            dense_block_host = np.empty((self.block_rows, num_cols), dtype=np.float32)
            dense_block_gpu = drv.mem_alloc(self.block_rows * num_cols * bytes_float)
            row_nnz_gpu = drv.mem_alloc(self.block_rows * np.dtype(np.int32).itemsize)

            block_size = 128

            # --- Count NNZ per row using GPU kernel to be consistent with filling logic ---
            for b in trange(0, num_rows, self.block_rows, desc='Comptage NNZ'):
                current_rows = min(self.block_rows, num_rows - b)
                # Fill dense_block_host from manip
                for r in range(current_rows):
                    global_row = b + r
                    n_idx = global_row // self.T
                    t_idx = global_row % self.T
                    dense_block_host[r, :] = self.manip.AcousticFields[n_idx].field[t_idx].flatten()
                drv.memcpy_htod(dense_block_gpu, dense_block_host)

                grid = ((current_rows + block_size - 1) // block_size, 1, 1)
                count_nnz_kernel(dense_block_gpu, row_nnz_gpu,
                                 np.int32(current_rows), np.int32(num_cols),
                                 np.float32(self.relative_threshold),
                                 block=(block_size, 1, 1), grid=grid)

                row_nnz_host = np.empty(current_rows, dtype=np.int32)
                drv.memcpy_dtoh(row_nnz_host, row_nnz_gpu)
                self.row_ptr[b + 1:b + current_rows + 1] = self.row_ptr[b] + np.cumsum(row_nnz_host, dtype=np.int64)

            # total nnz
            self.total_nnz = int(self.row_ptr[-1])
            print(f"NNZ total : {self.total_nnz}")

            # allocate final arrays
            self.h_col_ind = np.zeros(self.total_nnz, dtype=np.uint32)
            self.h_values = np.zeros(self.total_nnz, dtype=np.float32)

            # copy row_ptr to device once
            self.row_ptr_gpu = drv.mem_alloc(self.row_ptr.nbytes)
            drv.memcpy_htod(self.row_ptr_gpu, self.row_ptr)

            # allocate device arrays for final csr
            self.col_ind_gpu = drv.mem_alloc(self.h_col_ind.nbytes)
            self.values_gpu = drv.mem_alloc(self.h_values.nbytes)

            # --- Fill CSR per-block ---
            for b in trange(0, num_rows, self.block_rows, desc='Remplissage CSR'):
                current_rows = min(self.block_rows, num_rows - b)
                for r in range(current_rows):
                    global_row = b + r
                    n_idx = global_row // self.T
                    t_idx = global_row % self.T
                    dense_block_host[r, :] = self.manip.AcousticFields[n_idx].field[t_idx].flatten()
                drv.memcpy_htod(dense_block_gpu, dense_block_host)

                grid = ((current_rows + block_size - 1) // block_size, 1, 1)
                fill_csr_kernel(dense_block_gpu,
                                self.row_ptr_gpu,
                                self.col_ind_gpu,
                                self.values_gpu,
                                np.int32(b),
                                np.int32(current_rows),
                                np.int32(num_cols),
                                np.float32(self.relative_threshold),
                                np.int64(self.total_nnz),
                                block=(block_size, 1, 1), grid=grid)
                drv.Context.synchronize()

            # copy back
            drv.memcpy_dtoh(self.h_col_ind, self.col_ind_gpu)
            drv.memcpy_dtoh(self.h_values, self.values_gpu)
            print('CSR généré ✔')

            # compute normalization factor from CSR (sum per column)
            self.compute_norm_factor_from_csr()

            # free temporaries
            dense_block_gpu.free(); row_nnz_gpu.free()

        except Exception as e:
            print(f"❌ Erreur détaillée : {e}")
            self.free()
            raise

    def compute_norm_factor_from_csr(self):
        ZX = self.Z * self.X

        # 1) Allouer un vecteur de somme colonne sur le GPU
        col_sum_gpu = drv.mem_alloc(ZX * np.dtype(np.float32).itemsize)
        drv.memset_d32(col_sum_gpu, 0, ZX)

        # 2) Récupérer le kernel
        acc_kernel = self.sparse_mod.get_function("accumulate_columns_atomic__CSR")

        # 3) Lancer le kernel
        threads = 256
        blocks = (self.total_nnz + threads - 1) // threads

        acc_kernel(
            self.values_gpu,
            self.col_ind_gpu,
            np.int64(self.total_nnz),
            col_sum_gpu,
            block=(threads,1,1),
            grid=(blocks,1,1)
        )
        drv.Context.synchronize()

        # 4) Récupérer le résultat
        norm = np.empty(ZX, dtype=np.float32)
        drv.memcpy_dtoh(norm, col_sum_gpu)
        col_sum_gpu.free()

        norm = np.maximum(norm.astype(np.float64), 1e-6)
        self.norm_factor_inv = (1.0 / norm).astype(np.float32)

        self.norm_factor_inv_gpu = drv.mem_alloc(self.norm_factor_inv.nbytes)
        drv.memcpy_htod(self.norm_factor_inv_gpu, self.norm_factor_inv)

    def getMatrixSize(self):
        if self.row_ptr is None:
            return {"error": "La matrice sparse n'est pas encore allouée."}
        total = (self.row_ptr.nbytes if self.row_ptr is not None else 0) + \
                (self.h_col_ind.nbytes if self.h_col_ind is not None else 0) + \
                (self.h_values.nbytes if self.h_values is not None else 0)
        return total / (1024**3)

    def free(self):
        try:
            if hasattr(self, 'col_ind_gpu') and self.col_ind_gpu:
                self.col_ind_gpu.free()
            if hasattr(self, 'values_gpu') and self.values_gpu:
                self.values_gpu.free()
            if hasattr(self, 'row_ptr_gpu') and self.row_ptr_gpu:
                self.row_ptr_gpu.free()
            if hasattr(self, 'norm_factor_inv_gpu') and self.norm_factor_inv_gpu:
                self.norm_factor_inv_gpu.free()
            if hasattr(self, 'ctx') and self.ctx:
                try:
                    self.ctx.pop()
                except Exception:
                    pass
                self.ctx = None
            print('✅ Mémoire GPU libérée.')
        except Exception as e:
            print(f"❌ Erreur lors de la libération de la mémoire GPU : {e}")
        
    def compute_density(self):
        """
        Retourne la densité réelle de la CSR = NNZ / (num_rows * num_cols)
        Nécessite que self.h_values et self.row_ptr existent (host).
        """
        if self.row_ptr is None or self.h_values is None:
            raise RuntimeError("row_ptr et h_values requis pour calculer la densité")
        num_rows = int(self.N * self.T)
        num_cols = int(self.Z * self.X)
        total_nnz = int(self.row_ptr[-1])
        density = total_nnz / (num_rows * num_cols)
        return density
   


from AOT_biomaps.Config import config
import torch
import numpy as np
from tqdm import trange
from AOT_biomaps.AOT_Recon.ReconTools import calculate_memory_requirement, check_gpu_memory

def LS(
    SMatrix,
    y,
    numIterations=5000,
    alpha=1e-3,
    isSavingEachIteration=True,
    withTumor=True,
    device=None,
    max_saves=5000,
    show_logs=True
):
    """
    Least Squares reconstruction using Projected Gradient Descent (PGD) with non-negativity constraint.
    Currently only implements the stable GPU version.
    """
    tumor_str = "WITH" if withTumor else "WITHOUT"
    # Force GPU usage for now
    if device is None:
        if torch.cuda.is_available() and check_gpu_memory(config.select_best_gpu(), calculate_memory_requirement(SMatrix, y), show_logs=show_logs):
            raise RuntimeError("CUDA is required for this implementation.")
        device = torch.device(f"cuda:{config.select_best_gpu()}")
    else:
        if device.type != "cuda":
            raise RuntimeError("Only GPU implementation is available for now.")
    return _LS_GPU_stable(SMatrix, y, numIterations, alpha, isSavingEachIteration, tumor_str, max_saves, show_logs=show_logs)

def _LS_GPU_stable(SMatrix, y, numIterations, alpha, isSavingEachIteration, tumor_str, max_saves=5000, show_logs=True):
    """
    Stable GPU implementation of LS using projected gradient descent with diagonal preconditioner.
    """
    device = torch.device(f"cuda:{config.select_best_gpu()}")
    T, Z, X, N = SMatrix.shape
    ZX = Z * X
    TN = T * N
    # 1. Conversion et normalisation
    A_flat = torch.from_numpy(SMatrix).to(device=device, dtype=torch.float32).permute(0, 3, 1, 2).reshape(TN, ZX)
    y_flat = torch.from_numpy(y).to(device=device, dtype=torch.float32).reshape(TN)
    norm_A = A_flat.max()
    norm_y = y_flat.max()
    A_flat.div_(norm_A + 1e-8)
    y_flat.div_(norm_y + 1e-8)
    # 2. Initialisation
    lambda_k = torch.zeros(ZX, device=device)
    lambda_history = [] if isSavingEachIteration else None
    saved_indices = []  # Pour stocker les indices des itérations sauvegardées

    # Calculate save indices
    if numIterations <= max_saves:
        save_indices = list(range(numIterations))
    else:
        step = numIterations // max_saves
        save_indices = list(range(0, numIterations, step))
        if save_indices[-1] != numIterations - 1:
            save_indices.append(numIterations - 1)

    # Préconditionneur diagonal
    diag_AAT = torch.sum(A_flat ** 2, dim=0)
    M_inv = 1.0 / torch.clamp(diag_AAT, min=1e-6)
    # Pré-allocation des tenseurs
    r_k = torch.empty_like(y_flat)
    AT_r = torch.empty(ZX, device=device)
    description = f"AOT-BioMaps -- Stable LS Reconstruction ---- {tumor_str} TUMOR ---- GPU {torch.cuda.current_device()}"

    iterator = trange(numIterations, desc=description) if show_logs else range(numIterations)
    for it in iterator:
        # Calcul du résidu (inplace)
        torch.matmul(A_flat, lambda_k, out=r_k)
        r_k = y_flat - r_k
        if isSavingEachIteration and it in save_indices:
            lambda_history.append(lambda_k.clone().reshape(Z, X) * (norm_y / norm_A))
            saved_indices.append(it)

        # Gradient préconditionné (inplace)
        torch.matmul(A_flat.T, r_k, out=AT_r)
        AT_r *= M_inv
        # Mise à jour avec pas fixe et projection (inplace)
        lambda_k.add_(AT_r, alpha=alpha)
        lambda_k.clamp_(min=0)

    # 3. Dénormalisation
    lambda_final = lambda_k.reshape(Z, X) * (norm_y / norm_A)
    # Free memory
    del A_flat, y_flat, r_k, AT_r
    torch.cuda.empty_cache()
    if isSavingEachIteration:
        return [t.cpu().numpy() for t in lambda_history], saved_indices
    else:
        return lambda_final.cpu().numpy(), None

def _LS_GPU_opti(*args, **kwargs):
    raise NotImplementedError("Only _LS_GPU_stable is implemented for now.")

def _LS_GPU_multi(*args, **kwargs):
    raise NotImplementedError("Only _LS_GPU_stable is implemented for now.")

def _LS_CPU_opti(*args, **kwargs):
    raise NotImplementedError("Only _LS_GPU_stable is implemented for now.")

def _LS_CPU_basic(*args, **kwargs):
    raise NotImplementedError("Only _LS_GPU_stable is implemented for now.")



import os
import torch
import numpy as np
from numba import njit, prange
from torch_sparse import coalesce
import cupyx.scipy.sparse as cpsparse

def load_recon(hdr_path):
    """
    Lit un fichier Interfile (.hdr) et son fichier binaire (.img) pour reconstruire une image comme le fait Vinci.
    
    Paramètres :
    ------------
    - hdr_path : chemin complet du fichier .hdr
    
    Retour :
    --------
    - image : tableau NumPy contenant l'image
    - header : dictionnaire contenant les métadonnées du fichier .hdr
    """
    header = {}
    with open(hdr_path, 'r') as f:
        for line in f:
            if ':=' in line:
                key, value = line.split(':=', 1)  # s'assurer qu'on ne coupe que la première occurrence de ':='
                key = key.strip().lower().replace('!', '')  # Nettoyage des caractères
                value = value.strip()
                header[key] = value
    
    # 📘 Obtenez le nom du fichier de données associé (le .img)
    data_file = header.get('name of data file')
    if data_file is None:
        raise ValueError(f"Impossible de trouver le fichier de données associé au fichier header {hdr_path}")
    
    img_path = os.path.join(os.path.dirname(hdr_path), data_file)
    
    # 📘 Récupérer la taille de l'image à partir des métadonnées
    shape = [int(header[f'matrix size [{i}]']) for i in range(1, 4) if f'matrix size [{i}]' in header]
    if shape and shape[-1] == 1:  # Si la 3e dimension est 1, on la supprime
        shape = shape[:-1]  # On garde (192, 240) par exemple
    
    if not shape:
        raise ValueError("Impossible de déterminer la forme de l'image à partir des métadonnées.")
    
    # 📘 Déterminez le type de données à utiliser
    data_type = header.get('number format', 'short float').lower()
    dtype_map = {
        'short float': np.float32,
        'float': np.float32,
        'int16': np.int16,
        'int32': np.int32,
        'uint16': np.uint16,
        'uint8': np.uint8
    }
    dtype = dtype_map.get(data_type)
    if dtype is None:
        raise ValueError(f"Type de données non pris en charge : {data_type}")
    
    # 📘 Ordre des octets (endianness)
    byte_order = header.get('imagedata byte order', 'LITTLEENDIAN').lower()
    endianess = '<' if 'little' in byte_order else '>'
    
    # 📘 Vérifie la taille réelle du fichier .img
    img_size = os.path.getsize(img_path)
    expected_size = np.prod(shape) * np.dtype(dtype).itemsize
    
    if img_size != expected_size:
        raise ValueError(f"La taille du fichier img ({img_size} octets) ne correspond pas à la taille attendue ({expected_size} octets).")
    
    # 📘 Lire les données binaires et les reformater
    with open(img_path, 'rb') as f:
        data = np.fromfile(f, dtype=endianess + np.dtype(dtype).char)
    
    image =  data.reshape(shape[::-1]) 
    
    # 📘 Rescale l'image si nécessaire
    rescale_slope = float(header.get('data rescale slope', 1))
    rescale_offset = float(header.get('data rescale offset', 0))
    image = image * rescale_slope + rescale_offset
    
    return image

def mse(y_true, y_pred):
    """
    Calcule la Mean Squared Error (MSE) entre deux tableaux.
    Équivalent à sklearn.metrics.mean_squared_error.
    """
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)
    return np.mean((y_true - y_pred) ** 2)

def ssim(img1, img2, win_size=7, k1=0.01, k2=0.03, L=1.0):
    """
    Calcule l'SSIM entre deux images 2D (niveaux de gris).
    Équivalent à skimage.metrics.structural_similarity avec :
    - data_range=1.0 (images normalisées entre 0 et 1)
    - gaussian_weights=True (fenêtre gaussienne)
    - multichannel=False (1 canal)

    Args:
        img1, img2: Images 2D (numpy arrays) de même taille.
        win_size: Taille de la fenêtre de comparaison (doit être impair).
        k1, k2: Constantes pour stabiliser la division (valeurs typiques : 0.01, 0.03).
        L: Dynamique des pixels (1.0 si images dans [0,1], 255 si dans [0,255]).
    Returns:
        SSIM moyen sur l'image (float entre -1 et 1).
    """
    if img1.shape != img2.shape:
        raise ValueError("Les images doivent avoir la même taille.")
    if win_size % 2 == 0:
        raise ValueError("win_size doit être impair.")

    # Constantes
    C1 = (k1 * L) ** 2
    C2 = (k2 * L) ** 2

    # Fenêtre gaussienne
    window = np.ones((win_size, win_size)) / (win_size ** 2)  # Approximation (skimage utilise une gaussienne)
    window = window / np.sum(window)  # Normalisation

    # Pad les images pour éviter les bords
    pad = win_size // 2
    img1_pad = np.pad(img1, pad, mode='reflect')
    img2_pad = np.pad(img2, pad, mode='reflect')

    # Calcul des statistiques locales
    mu1 = np.zeros_like(img1, dtype=np.float64)
    mu2 = np.zeros_like(img1, dtype=np.float64)
    sigma1_sq = np.zeros_like(img1, dtype=np.float64)
    sigma2_sq = np.zeros_like(img1, dtype=np.float64)
    sigma12 = np.zeros_like(img1, dtype=np.float64)

    # Itère sur chaque pixel (optimisé avec des convolutions)
    for i in range(pad, img1_pad.shape[0] - pad):
        for j in range(pad, img1_pad.shape[1] - pad):
            patch1 = img1_pad[i-pad:i+pad+1, j-pad:j+pad+1]
            patch2 = img2_pad[i-pad:i+pad+1, j-pad:j+pad+1]

            mu1[i-pad, j-pad] = np.sum(patch1 * window)
            mu2[i-pad, j-pad] = np.sum(patch2 * window)
            sigma1_sq[i-pad, j-pad] = np.sum(window * (patch1 - mu1[i-pad, j-pad]) ** 2)
            sigma2_sq[i-pad, j-pad] = np.sum(window * (patch2 - mu2[i-pad, j-pad]) ** 2)
            sigma12[i-pad, j-pad] = np.sum(window * (patch1 - mu1[i-pad, j-pad]) * (patch2 - mu2[i-pad, j-pad]))

    # SSIM locale
    ssim_map = ((2 * mu1 * mu2 + C1) * (2 * sigma12 + C2)) / (
        (mu1**2 + mu2**2 + C1) * (sigma1_sq + sigma2_sq + C2)
    )

    return np.mean(ssim_map)

def calculate_memory_requirement(SMatrix, y):
    """
    Calcule la mémoire requise (en Go) pour :
    - SMatrix : np.ndarray (dense) ou cpsparse.csr_matrix (sparse)
    - y : vecteur (NumPy ou CuPy, float32)

    Args:
        SMatrix: np.ndarray ou cpsparse.csr_matrix
        y: vecteur (float32)
    """
    # 1. Mémoire pour SMatrix
    if isinstance(SMatrix, cpsparse.csr_matrix):
        # Matrice CSR CuPy
        nnz = SMatrix.nnz
        num_cols = SMatrix.shape[1]
        size_SMatrix = nnz * (4 + 4) + (num_cols + 1) * 4  # data (float32) + indices (int32) + indptr (int32)
    elif isinstance(SMatrix, np.ndarray):
        # Tableau NumPy dense (float32)
        size_SMatrix = SMatrix.nbytes
    else:
        raise ValueError("SMatrix doit être un np.ndarray ou une matrice CSR CuPy.")

    # 2. Mémoire pour y (float32)
    if hasattr(y, 'nbytes'):
        size_y = y.nbytes
    else:
        size_y = np.prod(y.shape) * 4  # float32

    # Total en Go
    total_bytes = size_SMatrix + size_y
    total_GB = total_bytes / 1024**3

    return total_GB


def check_gpu_memory(device_index, required_memory, show_logs=True):
    """Check if enough memory is available on the specified GPU."""
    free_memory, _ = torch.cuda.mem_get_info(f"cuda:{device_index}")
    free_memory_gb = free_memory / 1024**3
    if show_logs:
        print(f"Free memory on GPU {device_index}: {free_memory_gb:.2f} GB, Required memory: {required_memory:.2f} GB")
    return free_memory_gb >= required_memory

@njit(parallel=True)
def _forward_projection(SMatrix, theta_p, q_p):
    t_dim, z_dim, x_dim, i_dim = SMatrix.shape
    for _t in prange(t_dim):
        for _n in range(i_dim):
            total = 0.0
            for _z in range(z_dim):
                for _x in range(x_dim):
                    total += SMatrix[_t, _z, _x, _n] * theta_p[_z, _x]
            q_p[_t, _n] = total

@njit(parallel=True)
def _backward_projection(SMatrix, e_p, c_p):
    t_dim, z_dim, x_dim, n_dim = SMatrix.shape
    for _z in prange(z_dim):
        for _x in range(x_dim):
            total = 0.0
            for _t in range(t_dim):
                for _n in range(n_dim):
                    total += SMatrix[_t, _z, _x, _n] * e_p[_t, _n]
            c_p[_z, _x] = total


def _build_adjacency_sparse(Z, X, device, corner=(0.5 - np.sqrt(2) / 4) / np.sqrt(2), face=0.5 - np.sqrt(2) / 4,dtype=torch.float32):
    rows, cols, weights = [], [], []
    for z in range(Z):
        for x in range(X):
            j = z * X + x
            for dz, dx in [(-1, -1), (-1, 0), (-1, 1),
                           (0, -1),           (0, 1),
                           (1, -1),   (1, 0), (1, 1)]:
                nz, nx = z + dz, x + dx
                if 0 <= nz < Z and 0 <= nx < X:
                    k = nz * X + nx
                    weight = corner if abs(dz) + abs(dx) == 2 else face
                    rows.append(j)
                    cols.append(k)
                    weights.append(weight)
    index = torch.tensor([rows, cols], dtype=torch.long, device=device)
    values = torch.tensor(weights, dtype=dtype, device=device)
    index, values = coalesce(index, values, m=Z*X, n=Z*X)
    return index, values


def power_method(P, PT, data, Z, X, n_it=10):
    x = torch.randn(Z * X, device=data.device)
    x = x / torch.norm(x)
    for _ in range(n_it):
        Ax = P(x)
        ATax = PT(Ax)
        x = ATax / torch.norm(ATax)
    ATax = PT(P(x))
    return torch.sqrt(torch.dot(x, ATax))

def proj_l2(p, alpha):
    if alpha <= 0:
        return torch.zeros_like(p)
    norm = torch.sqrt(torch.sum(p**2, dim=0, keepdim=True) + 1e-12)
    return p * torch.min(norm, torch.tensor(alpha, device=p.device)) / (norm + 1e-12)

def gradient(x):
    grad_x = torch.zeros_like(x)
    grad_y = torch.zeros_like(x)
    grad_x[:, :-1] = x[:, 1:] - x[:, :-1]  # Gradient horizontal
    grad_y[:-1, :] = x[1:, :] - x[:-1, :]   # Gradient vertical
    return torch.stack((grad_x, grad_y), dim=0)

def div(x):
    if x.dim() == 3:
        x = x.unsqueeze(0)  # Ajoute une dimension batch si nécessaire

    gx = x[:, 0, :, :]  # Gradient horizontal (shape: [1, H, W] ou [H, W])
    gy = x[:, 1, :, :]  # Gradient vertical   (shape: [1, H, W] ou [H, W])

    # Divergence du gradient horizontal (gx)
    div_x = torch.zeros_like(gx)
    div_x[:, :, 1:] += gx[:, :, :-1]  # Contribution positive (gauche)
    div_x[:, :, :-1] -= gx[:, :, :-1] # Contribution négative (droite)

    # Divergence du gradient vertical (gy)
    div_y = torch.zeros_like(gy)
    div_y[:, 1:, :] += gy[:, :-1, :]  # Contribution positive (haut)
    div_y[:, :-1, :] -= gy[:, :-1, :] # Contribution négative (bas)

    return -(div_x + div_y)

def norm2sq(x):
    return torch.sum(x**2)

def norm1(x):
    return torch.sum(torch.abs(x))

def KL_divergence(Ax, y):
    return torch.sum(Ax - y * torch.log(Ax + 1e-10))

def gradient_KL(Ax, y):
    return 1 - y / (Ax + 1e-10)

def prox_F_star(y, sigma, a):
    return 0.5 * (y - torch.sqrt(y**2 + 4 * sigma * a))

def prox_G(x, tau, K):
    return torch.clamp(x - tau * K, min=0)

def filter_radon(f, N, filter_type, Fc):
    """
    Implémente les filtres pour la rétroprojection filtrée (iRadon).
    Inspirée de la fonction MATLAB FilterRadon de Mamouna Bocoum.

    Paramètres :
    ------------
    f : np.ndarray
        Vecteur des fréquences (ex: f_t ou f_z).
    N : int
        Taille du filtre (longueur de f).
    filter_type : str
        Type de filtre : 'ram-lak', 'shepp-logan', 'cosine', 'hamming', 'hann'.
    Fc : float
        Fréquence de coupure.

    Retourne :
    -----------
    FILTER : np.ndarray
        Filtre appliqué aux fréquences.
    """
    FILTER = np.abs(f)

    if filter_type == 'ram-lak':
        pass  # FILTER = |f| (déjà calculé)
    elif filter_type == 'shepp-logan':
        # Évite la division par zéro
        with np.errstate(divide='ignore', invalid='ignore'):
            FILTER = FILTER * (np.sinc(2 * f / (2 * Fc)))  # sin(2πf/(2Fc))/(2πf/(4Fc)) = sinc(2f/(2Fc))
        FILTER[np.isnan(FILTER)] = 1.0  # Pour f=0
    elif filter_type == 'cosine':
        FILTER = FILTER * np.cos(2 * np.pi * f / (4 * Fc))
    elif filter_type == 'hamming':
        FILTER = FILTER * (0.54 + 0.46 * np.cos(2 * np.pi * f / Fc))
    elif filter_type == 'hann':
        FILTER = FILTER * (1 + np.cos(2 * np.pi * f / (4 * Fc))) / 2
    else:
        raise ValueError(f"Type de filtre inconnu : {filter_type}")

    # Coupure des fréquences au-delà de Fc
    FILTER[np.abs(f) > Fc] = 0
    # Atténuation exponentielle (optionnelle, comme dans le code MATLAB)
    FILTER = FILTER * np.exp(-2 * (np.abs(f) / Fc)**10)

    return FILTER



from AOT_biomaps.Config import config

from scipy.signal import hilbert
import os
import h5py
import numpy as np
import torch
from concurrent.futures import ThreadPoolExecutor

def loadmat(param_path_mat):
    """
    Charge un fichier .mat (format HDF5) sans SciPy.
    Args:
        param_path_mat: Chemin vers le fichier .mat.
    Returns:
        Dictionnaire contenant les variables du fichier.
    """
    with h5py.File(param_path_mat, 'r') as f:
        data = {}
        for key in f.keys():
            # Récupère les données et convertit en numpy array si nécessaire
            item = f[key]
            if isinstance(item, h5py.Dataset):
                data[key] = item[()]  # Convertit en numpy array
            elif isinstance(item, h5py.Group):
                # Pour les structures MATLAB (nested)
                data[key] = {}
                for subkey in item:
                    data[key][subkey] = item[subkey][()]
    return data

def reshape_field(field, factor, device=None):
    """
    Downsample a 3D or 4D field using PyTorch interpolation (auto-detects GPU/CPU).
    Args:
        field: Input field (numpy array or torch.Tensor).
        factor: Downsampling factor (tuple of ints).
        device: Force device ('cpu' or 'cuda'). If None, auto-detects GPU.
    Returns:
        Downsampled field (same type as input: numpy array or torch.Tensor).
    """
    # Check input
    if field is None:
        raise ValueError("Acoustic field is not generated. Please generate the field first.")

    # Auto-detect device if not specified
    if device is None:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
    else:
        device = device.lower()
        if device not in ['cpu', 'cuda']:
            raise ValueError("Device must be 'cpu' or 'cuda'.")

    # Convert to torch.Tensor if needed
    if isinstance(field, np.ndarray):
        field = torch.from_numpy(field)
    elif not isinstance(field, torch.Tensor):
        raise TypeError("Input must be a numpy array or torch.Tensor.")

    # Move to the target device
    field = field.to(device)

    # Add batch and channel dimensions (required by torch.interpolate)
    if len(factor) == 3:
        if field.dim() != 3:
            raise ValueError("Expected 3D field.")
        field = field.unsqueeze(0).unsqueeze(0)  # (1, 1, D, H, W)

        # Calculate new shape
        new_shape = [
            field.shape[2] // factor[0],
            field.shape[3] // factor[1],
            field.shape[4] // factor[2]
        ]

        # Trilinear interpolation
        downsampled = torch.nn.functional.interpolate(
            field,
            size=new_shape,
            mode='trilinear',
            align_corners=True
        )
        downsampled = downsampled.squeeze(0).squeeze(0)  # Remove batch/channel dims

    elif len(factor) == 4:
        if field.dim() != 4:
            raise ValueError("Expected 4D field.")
        field = field.unsqueeze(0).unsqueeze(0)  # (1, 1, T, D, H, W)

        new_shape = [
            field.shape[2] // factor[0],
            field.shape[3] // factor[1],
            field.shape[4] // factor[2],
            field.shape[5] // factor[3]
        ]

        # Tetra-linear interpolation
        downsampled = torch.nn.functional.interpolate(
            field,
            size=new_shape,
            mode='trilinear',  # PyTorch uses 'trilinear' for both 3D and 4D
            align_corners=True
        )
        downsampled = downsampled.squeeze(0).squeeze(0)

    else:
        raise ValueError("Unsupported dimension. Only 3D and 4D fields are supported.")

    # Convert back to numpy if input was numpy
    if isinstance(field, np.ndarray):
        return downsampled.cpu().numpy()
    else:
        return downsampled.cpu().numpy()

def calculate_envelope_squared(field):
    """
    Calcule l'enveloppe au carré du champ acoustique en utilisant scipy.signal.hilbert (CPU uniquement).

    Args:
        field: Champ acoustique (numpy.ndarray) de forme (T, X, Z) ou (T, X, Y, Z).

    Returns:
        envelope (numpy.ndarray): Enveloppe au carré du champ acoustique.
    """
    try:
        if field is None:
            raise ValueError("Le champ acoustique n'est pas généré. Veuillez d'abord générer le champ.")

        if not isinstance(field, np.ndarray):
            if hasattr(field, 'cpu'):
                field = field.cpu().numpy()  # Si c'est un tenseur PyTorch sur GPU/CPU
            else:
                field = np.array(field)  # Conversion générique

        if len(field.shape) not in [3, 4]:
            raise ValueError("Le champ acoustique doit être un tableau 3D (T, X, Z) ou 4D (T, X, Y, Z).")

        # Calcul de l'enveloppe avec scipy.signal.hilbert
        if len(field.shape) == 3:
            T, X, Z = field.shape
            envelope = np.zeros_like(field)
            for x in range(X):
                for z in range(Z):
                    envelope[:, x, z] = np.abs(hilbert(field[:, x, z], axis=0)) ** 2
        elif len(field.shape) == 4:
            T, X, Y, Z = field.shape
            envelope = np.zeros_like(field)
            for x in range(X):
                for y in range(Y):
                    for z in range(Z):
                        envelope[:, x, y, z] = np.abs(hilbert(field[:, x, y, z], axis=0)) ** 2
                        

        return envelope

    except Exception as e:
        print(f"Erreur dans calculate_envelope_squared: {e}")
        raise



def getPattern(pathFile):
    """
    Get the pattern from a file path.

    Args:
        pathFile (str): Path to the file containing the pattern.

    Returns:
        str: The pattern string.
    """
    try:
        # Pattern between first _ and last _
        pattern = os.path.basename(pathFile).split('_')[1:-1]
        pattern_str = ''.join(pattern)
        return pattern_str
    except Exception as e:
        print(f"Error reading pattern from file: {e}")
        return None
    
def detect_space_0_and_space_1(hex_string):
    binary_string = bin(int(hex_string, 16))[2:].zfill(len(hex_string) * 4)
    
    # Trouver la plus longue séquence de 0 consécutifs
    zeros_groups = [len(s) for s in binary_string.split('1')]
    space_0 = max(zeros_groups) if zeros_groups else 0

    # Trouver la plus longue séquence de 1 consécutifs
    ones_groups = [len(s) for s in binary_string.split('0')]
    space_1 = max(ones_groups) if ones_groups else 0

    return space_0, space_1

def getAngle(pathFile):
    """
    Get the angle from a file path.

    Args:
        pathFile (str): Path to the file containing the angle.

    Returns:
        int: The angle in degrees.
    """
    try:
        # Angle between last _ and .
        angle_str = os.path.basename(pathFile).split('_')[-1].replace('.', '')
        if angle_str.startswith('0'):
            angle_str = angle_str[1:]
        elif angle_str.startswith('1'):
            angle_str = '-' + angle_str[1:]
        else:
            raise ValueError("Invalid angle format in file name.")
        return int(angle_str)
    except Exception as e:
        print(f"Error reading angle from file: {e}")
        return None

def next_power_of_2(n):
    """Calculate the next power of 2 greater than or equal to n."""
    return int(2 ** np.ceil(np.log2(n)))
        



from AOT_biomaps.AOT_Recon.ReconEnums import PotentialType
from AOT_biomaps.AOT_Recon.ReconTools import _build_adjacency_sparse, calculate_memory_requirement, check_gpu_memory
from AOT_biomaps.Config import config

import warnings
import numpy as np
import torch
from tqdm import trange

if config.get_process() == 'gpu':
    try:
        from torch_scatter import scatter
    except ImportError:
        raise ImportError("torch_scatter and torch_sparse are required for GPU processing. Please install them using 'pip install torch-scatter torch-sparse' with correct link (follow instructions https://github.com/LucasDuclos/AcoustoOpticTomography/edit/main/README.md).")

def DEPIERRO(
        SMatrix,
        y, 
        numIterations,
        beta,
        sigma,
        isSavingEachIteration,
        withTumor, 
        max_saves, 
        show_logs):
        """
        This method implements the DEPIERRO algorithm using either CPU or single-GPU PyTorch acceleration.
        Multi-GPU and Multi-CPU modes are not implemented for this algorithm.
        """
        try:
            tumor_str = "WITH" if withTumor else "WITHOUT"
            # Auto-select device and method
            if device is None:
                if torch.cuda.is_available() and check_gpu_memory(config.select_best_gpu(), calculate_memory_requirement(SMatrix, y), show_logs=show_logs):
                    device = torch.device(f"cuda:{config.select_best_gpu()}")
                    use_gpu = True
                else:
                    device = torch.device("cpu")
                    use_gpu = False
            else:
                use_gpu = device.type == "cuda"
            # Dispatch to the appropriate implementation
            if use_gpu:
                return _DEPIERRO_GPU(SMatrix, y, numIterations, beta, sigma, isSavingEachIteration, tumor_str, device, max_saves, show_logs)
            else:
                return _DEPIERRO_CPU(SMatrix, y, numIterations, beta, sigma, isSavingEachIteration, tumor_str, device, max_saves, show_logs)
        except Exception as e:
            print(f"Error in MLEM: {type(e).__name__}: {e}")
            return None, None

def _DEPIERRO_GPU(SMatrix, y, numIterations, beta, sigma, isSavingEachIteration, tumor_str, device, max_saves, show_logs=True):
    # Conversion des données en tenseurs PyTorch (float64)
    A_matrix_torch = torch.tensor(SMatrix, dtype=torch.float64, device=device)
    y_torch = torch.tensor(y, dtype=torch.float64, device=device)
    # Dimensions
    T, Z, X, N = SMatrix.shape
    J = Z * X
    # Redimensionnement des matrices
    A_flat = A_matrix_torch.permute(0, 3, 1, 2).reshape(T * N, J)
    y_flat = y_torch.reshape(-1)
    # Initialisation de theta
    theta_0 = torch.ones((Z, X), dtype=torch.float64, device=device)
    matrix_theta_torch = [theta_0.clone()]  # Clone pour éviter les références
    I_reconMatrix = [theta_0.cpu().numpy()]
    # Facteur de normalisation
    normalization_factor = A_matrix_torch.sum(dim=(0, 3))
    normalization_factor_flat = normalization_factor.reshape(-1)
    # Construction de la matrice d'adjacence
    adj_index, adj_values = _build_adjacency_sparse(Z, X, device=device, dtype=torch.float64)
    # Description pour la barre de progression
    description = f"AOT-BioMaps -- Bayesian Reconstruction Tomography: DE PIERRO (Sparse QUADRATIC β:{beta:.4f}, σ:{sigma:.4f}) ---- {tumor_str} TUMOR ---- processing on single GPU no.{torch.cuda.current_device()}"
    # Configuration pour la sauvegarde des itérations
    saved_indices = [0]

    # Calculate save indices
    if numIterations <= max_saves:
        save_indices = list(range(numIterations))
    else:
        step = numIterations // max_saves
        save_indices = list(range(0, numIterations, step))
        if save_indices[-1] != numIterations - 1:
            save_indices.append(numIterations - 1)

    # Boucle principale MAP-EM
    iterator = trange(numIterations, desc=description) if show_logs else range(numIterations)
    for it in iterator:
        theta_p = matrix_theta_torch[-1]
        theta_p_flat = theta_p.reshape(-1)
        # Étape 1 : Projection avant
        q_flat = A_flat @ theta_p_flat
        q_flat = q_flat + torch.finfo(torch.float64).tiny  # Évite la division par zéro
        # Étape 2 : Estimation de l'erreur
        e_flat = y_flat / q_flat
        # Étape 3 : Rétroprojection de l'erreur
        c_flat = A_flat.T @ e_flat
        # Étape 4 : Mise à jour multiplicative (EM)
        theta_EM_p_flat = theta_p_flat * c_flat
        # Étape 5 : Calcul de W_j et gamma_j
        W_j = scatter(adj_values, adj_index[0], dim=0, dim_size=J, reduce='sum') * (1.0 / (sigma**2))
        theta_k = theta_p_flat[adj_index[1]]
        weighted_theta_k = theta_k * adj_values
        gamma_j = theta_p_flat * W_j + scatter(weighted_theta_k, adj_index[0], dim=0, dim_size=J, reduce='sum')
        # Étape 6 : Mise à jour de De Pierro (résolution quadratique)
        A_coeff = 2 * beta * W_j
        B = -beta * gamma_j + normalization_factor_flat
        C = -theta_EM_p_flat
        discriminant = B**2 - 4 * A_coeff * C
        discriminant = torch.clamp(discriminant, min=0)
        theta_p_plus_1_flat = (-B + torch.sqrt(discriminant)) / (2 * A_coeff + torch.finfo(torch.float64).tiny)
        theta_p_plus_1_flat = torch.clamp(theta_p_plus_1_flat, min=0)
        # Étape 7 : Mise à jour de theta
        theta_next = theta_p_plus_1_flat.reshape(Z, X)
        matrix_theta_torch.append(theta_next)  # Ajoute la nouvelle itération
        # Sauvegarde conditionnelle
        if isSavingEachIteration and it in save_indices:
            I_reconMatrix.append(theta_next.cpu().numpy())
            saved_indices.append(it)
        # Libération mémoire partielle (optionnel, à ajuster selon besoin)
        del theta_p_flat, q_flat, e_flat, c_flat, theta_EM_p_flat, theta_p_plus_1_flat
        torch.cuda.empty_cache()

    # Libération finale des tenseurs GPU
    del A_matrix_torch, y_torch, A_flat, y_flat, normalization_factor, normalization_factor_flat
    torch.cuda.empty_cache()
    # Retour du résultat
    if isSavingEachIteration:
        return I_reconMatrix, saved_indices
    else:
        return matrix_theta_torch[-1].cpu().numpy(), None

def _DEPIERRO_CPU(SMatrix, y, numIterations, beta, sigma, isSavingEachIteration, tumor_str, device, max_saves, show_logs=True):
    try:
        if beta is None or sigma is None:
            raise ValueError("Depierro95 optimizer requires beta and sigma parameters.")

        A_matrix = np.array(SMatrix, dtype=np.float32)
        y_array = np.array(y, dtype=np.float32)
        T, Z, X, N = SMatrix.shape
        J = Z * X
        A_flat = A_matrix.transpose(0, 3, 1, 2).reshape(T * N, Z * X)
        y_flat = y_array.reshape(-1)
        theta_0 = np.ones((Z, X), dtype=np.float32)
        matrix_theta = [theta_0]
        I_reconMatrix = [theta_0.copy()]
        saved_indices = [0]
        normalization_factor = A_matrix.sum(axis=(0, 3))
        normalization_factor_flat = normalization_factor.reshape(-1)
        adj_index, adj_values = _build_adjacency_sparse(Z, X)

        # Calculate save indices
        if numIterations <= max_saves:
            save_indices = list(range(numIterations))
        else:
            step = numIterations // max_saves
            save_indices = list(range(0, numIterations, step))
            if save_indices[-1] != numIterations - 1:
                save_indices.append(numIterations - 1)

        description = f"AOT-BioMaps -- Bayesian Reconstruction Tomography: DE PIERRO (Sparse QUADRATIC β:{beta:.4f}, σ:{sigma:.4f}) ---- {tumor_str} TUMOR ---- processing on single CPU ----"

        iterator = trange(numIterations, desc=description) if show_logs else range(numIterations)
        for it in iterator:
            theta_p = matrix_theta[-1]
            theta_p_flat = theta_p.reshape(-1)
            q_flat = np.dot(A_flat, theta_p_flat)
            e_flat = y_flat / (q_flat + np.finfo(np.float32).tiny)
            c_flat = np.dot(A_flat.T, e_flat)
            theta_EM_p_flat = theta_p_flat * c_flat
            alpha_j = normalization_factor_flat
            W_j = np.bincount(adj_index[0], weights=adj_values, minlength=J) * (1.0 / sigma**2)
            theta_k = theta_p_flat[adj_index[1]]
            weighted_theta_k = theta_k * adj_values
            gamma_j = theta_p_flat * W_j + np.bincount(adj_index[0], weights=weighted_theta_k, minlength=J)
            A = 2 * beta * W_j
            B = -beta * gamma_j + alpha_j
            C = -theta_EM_p_flat
            theta_p_plus_1_flat = (-B + np.sqrt(B**2 - 4 * A * C)) / (2 * A + np.finfo(np.float32).tiny)
            theta_p_plus_1_flat = np.clip(theta_p_plus_1_flat, a_min=0, a_max=None)
            theta_next = theta_p_plus_1_flat.reshape(Z, X)
            matrix_theta[-1] = theta_next
            if isSavingEachIteration and it in save_indices:
                I_reconMatrix.append(theta_next.copy())
                saved_indices.append(it)

        if isSavingEachIteration:
            return I_reconMatrix, saved_indices
        else:
            return I_reconMatrix[-1], None
    except Exception as e:
        print(f"An error occurred in _DEPIERRO_CPU: {e}")
        return None, None
